/*
  MPIO block device driver, version 0.0.1
  Copyright (C) 2002 Yury Umanets (umka@namesys.com)

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, version 2.
*/

#include <linux/config.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/usb.h>
#include <asm/uaccess.h>

#include "mpio.h"
#include "device.h"

#define DEVICE_NAME         "mpio"
#define THREAD_NAME         "mpiod"
#define MPIO_BLOCK_MAJOR    40
#define MAJOR_NR            MPIO_BLOCK_MAJOR
#define DEVICE_REQUEST      mpioblock_request

#define DEVICE_NR(device)   (device)
#define DEVICE_ON(device)
#define DEVICE_OFF(device)
#define DEVICE_NO_RANDOM
#include <linux/blk.h>

#ifndef QUEUE_EMPTY
#define QUEUE_EMPTY (!CURRENT)
#endif

#define QUEUE_PLUGGED (blk_dev[MAJOR_NR].request_queue.plugged)

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
static devfs_handle_t mpio_devfs_handle = NULL;
#endif

#define DRIVER_VERSION "0.0.1"
#define DRIVER_AUTHOR "Yury Umanets (umka@namesys.com)"
#define DRIVER_DESC "MPIO block device driver"

#define MPIO_VENDOR_ID 0x2735
#define MPIO_PRODUCT_ID	0x1

/* Table of devices that work with this driver */
static struct usb_device_id mpioblock_table[] = {
	{ USB_DEVICE(MPIO_VENDOR_ID, MPIO_PRODUCT_ID) },
	{ }
};

MODULE_DEVICE_TABLE(usb, mpioblock_table);

#define MPIO_MAX_DEVICES 256

static int mpio_sizes[MPIO_MAX_DEVICES];
static int mpio_blksizes[MPIO_MAX_DEVICES];

/* Structure to hold all of our device specific stuff */
/* Array of pointers to our devices that are currently connected */
static struct mpioblock_dev *minor_table[MPIO_MAX_DEVICES];

/* Lock to protect the minor_table structure */
static DECLARE_MUTEX(minor_table_mutex);

static inline void mpioblock_free(struct mpioblock_dev *dev);

/* These function serve for safe gettin/putting sizes, and minor
 * devices */
static inline int mpioblock_get_size(int minor) {
	int size;
	
	down(&minor_table_mutex);
	size = mpio_sizes[minor];
	up(&minor_table_mutex);
	
	return size;
}

static inline void mpioblock_put_size(int minor, int size) {
	down(&minor_table_mutex);
	mpio_sizes[minor] = size;
	up(&minor_table_mutex);
}

static inline int mpioblock_get_blksize(int minor) {
	int size;
	
	down(&minor_table_mutex);
	size = mpio_blksizes[minor];
	up(&minor_table_mutex);
	
	return size;
}

static inline void mpioblock_put_blksize(int minor, int size) {
	down(&minor_table_mutex);
	mpio_blksizes[minor] = size;
	up(&minor_table_mutex);
}

static inline struct mpioblock_dev *table_get_device(int minor) {
	struct mpioblock_dev *dev;
	
	down(&minor_table_mutex);
	dev = minor_table[minor];
	up(&minor_table_mutex);
	
	return dev;
}

static inline void table_put_device(int minor, struct mpioblock_dev *dev) {
	down(&minor_table_mutex);
	minor_table[minor] = dev;
	up(&minor_table_mutex);
}

static inline int table_find_free(void) {
	int minor = 0;
	
	down(&minor_table_mutex);
	
	for (minor = 0; minor < MPIO_MAX_DEVICES; minor++) {
		if (minor_table[minor] == NULL)
			break;
	}
	
	up(&minor_table_mutex);
	
	return minor;
}

static int mpioblock_open(struct inode *inode, struct file *file) {
	int minor, retval = 0;
	struct mpioblock_dev *dev;
	
	minor = MINOR(inode->i_rdev);
	
	if (minor < 0 || minor >= MPIO_MAX_DEVICES)
		return -ENODEV;

	MOD_INC_USE_COUNT;

	if ((dev = table_get_device(minor)) == NULL) {
		MOD_DEC_USE_COUNT;
		return -ENODEV;
	}

	/* Setting up the device. The main thing is to attach it to file private
	 * data and let us be able to get it from there in the right time. Also
	 * we initialize mpio here */
	down(&dev->sem);
	
	if ((retval = mpio_init(dev->mpio))) {
		err("Can't initialize mpio device.");
		MOD_DEC_USE_COUNT;
		goto error_up_sem;
	}
	
	mpioblock_put_size(dev->minor, dev->mpio->size >> 10);
	mpioblock_put_blksize(dev->minor, 512);
	
	dev->counter++;
	file->private_data = dev;

 error_up_sem:
	up(&dev->sem);

	return retval;
}

static int mpioblock_release(struct inode *inode, struct file *file) {
	int retval = 0;
	struct mpioblock_dev *dev;

	if (inode == NULL)
		return -ENODEV;

	if (!(dev = table_get_device(MINOR(inode->i_rdev))))
		return -ENODEV;

	/* Lock our device */
	down(&dev->sem);

	if (dev->counter <= 0) {
		retval = -ENODEV;
		goto exit_not_freed;
	}

	/* Check if the device was unplugged before the file was released */
	if (dev->mpio == NULL) {
		up(&dev->sem);
		
		info("Device was unplugged before file was released.");
		
		mpioblock_free(dev);
		MOD_DEC_USE_COUNT;
		return 0;
	}

	/* Decrement our usage count for the device */
	dev->counter--;
	
	/* Shutdown any bulk writes that might be going on */
	if (dev->counter <= 0) {
		mpio_fini(dev->mpio);
		dev->counter = 0;
	}

	/* Decrement our usage count for the module */
	MOD_DEC_USE_COUNT;
	
 exit_not_freed:
	
	up(&dev->sem);
	return retval;
}

static int mpioblock_ioctl(struct inode *inode, struct file *file,
			   unsigned int cmd, unsigned long arg)
{
	int retval = 0;
	struct mpioblock_dev *dev;

	if (inode == NULL)
		return -ENODEV;

	if (!(dev = table_get_device(MINOR(inode->i_rdev))))
		return -ENODEV;

	switch (cmd) {
	case BLKGETSIZE:
		down(&dev->sem);
		
		retval = put_user((dev->mpio->size >> 9),
				  (unsigned long *)arg);
		
		up(&dev->sem);

		return retval;
	case BLKFLSBUF:
		if (!capable(CAP_SYS_ADMIN))
			return -EACCES;
		
		/* FIXME: Buffers will be flushed here when buffering ready */
		
		fsync_dev(inode->i_rdev);
		invalidate_buffers(inode->i_rdev);

		return 0;
	default:
		return -EINVAL;
	}
}

static struct block_device_operations mpio_fops = {
	.owner      = THIS_MODULE,
	.open       = mpioblock_open,
	.release    = mpioblock_release,
	.ioctl      = mpioblock_ioctl
};      

static inline struct mpioblock_dev *mpioblock_create(struct usb_device *udev,
						     int ifnum)
{
	int minor;
	char name[10];
	struct mpioblock_dev *dev;

	if ((minor = table_find_free()) >= MPIO_MAX_DEVICES) {
		info("Too many devices plugged in. Can't handle this device.");
		return NULL;
	}

	/* Allocate memory for our device state and intialize it */
	if (!(dev = kmalloc(sizeof(struct mpioblock_dev), GFP_KERNEL))) {
		err("Out of memory");
		return NULL;
	}

	memset(dev, 0, sizeof(*dev));

	init_MUTEX(&dev->sem);
	dev->minor = minor;

	if (!(dev->mpio = mpio_create(udev, ifnum))) {
		err("Can't create mpio device.");
		goto error_free_dev;
	}

#ifdef CONFIG_DEVFS_FS
	sprintf(name, "mpio%d", minor);
	
	dev->devfs = devfs_register(mpio_devfs_handle, name, DEVFS_FL_DEFAULT,
				    MPIO_BLOCK_MAJOR, minor, S_IFBLK |
				    S_IRUGO | S_IWUGO, &mpio_fops, NULL);
#endif
	
	table_put_device(minor, dev);
	info("MPIO device attached to #%d", minor);

	return dev;

 error_free_dev:
	kfree(dev);
	return NULL;
}

/* Freeing device and removing it from the minor table */
static inline void mpioblock_free(struct mpioblock_dev *dev) {
	down(&dev->sem);
	
	table_put_device(dev->minor, NULL);

	if (dev->mpio)
		mpio_free(dev->mpio);

	up(&dev->sem);
	kfree(dev);
}

/*
  Called by the usb core when a new device is connected that it thinks this
  driver might be interested in.
*/
static void *mpioblock_probe(struct usb_device *udev, unsigned ifnum,
			     const struct usb_device_id *id)
{
	struct mpioblock_dev *dev;
	
	/* See if the device offered us matches what we can accept */
	if ((udev->descriptor.idVendor != MPIO_VENDOR_ID) ||
	    (udev->descriptor.idProduct != MPIO_PRODUCT_ID))
	{
		return NULL;
	}

	if (!(dev = mpioblock_create(udev, ifnum))) {
		err("Can't create new mpio device.");
		return NULL;
	}

	return dev;
}

static void mpioblock_disconnect(struct usb_device *udev, void *ptr) {
	struct mpioblock_dev *dev;

	dev = (struct mpioblock_dev *)ptr;

	down(&dev->sem);

#ifdef CONFIG_DEVFS_FS
	devfs_unregister(dev->devfs);
#endif

	mpioblock_put_size(dev->minor, 0);
	mpioblock_put_blksize(dev->minor, 0);
	
	if (!dev->counter) {
		up(&dev->sem);
		mpioblock_free(dev);
	} else {
		mpio_free(dev->mpio);
		dev->mpio = NULL;
		up(&dev->sem);
	}

	info("MPIO #%d detached", dev->minor);
}

/* USB specific object needed to register this driver with the usb subsystem */
static struct usb_driver mpioblock_driver = {
	.name       = "mpio",
	.probe      = mpioblock_probe,
	.disconnect = mpioblock_disconnect,
	.id_table   = mpioblock_table,
};

static int mpioblock_io(struct mpioblock_dev *dev, int cmd,
			void *buff, int sector, int count)
{
	int retval, i;
	
	down(&dev->sem);

	if (dev->mpio == NULL) {
		up(&dev->sem);
		return -EINTR;
	}

	for (i = 0; i < count; i++) {
		
		if (dev->mpio == NULL) {
			up(&dev->sem);
			return -EINTR;
		}

		if (cmd == READ) {
			if ((retval = mpio_read(dev->mpio, buff, sector + i)))
				break;
		} else {
			if ((retval = mpio_write(dev->mpio, buff, sector + i)))
				break;
		}
		
		buff += SECTOR_SIZE;
	}

	up(&dev->sem);

	return retval;
}

static void mpioblock_handle_request(void) {
	int retval;
	struct request *req;
	struct mpioblock_dev *dev;

	for (;;) {
		INIT_REQUEST;
		req = CURRENT;
		spin_unlock_irq(&io_request_lock);
		
		retval = 0;
		
		if (!(dev = table_get_device(MINOR(req->rq_dev)))) {
			err("Invalid device detected while handling IO request.");
			goto end_request;
		}

		if (dev->mpio == NULL) {
			err("Device unplugged while handling IO request.");
			goto end_request;
		}
		
		if (MINOR(req->rq_dev) >= MPIO_MAX_DEVICES)
			panic(__FUNCTION__": minor is out of bound");

		if ((req->sector + req->current_nr_sectors) <= (dev->mpio->size >> 9)) {
			retval = !mpioblock_io(dev, req->cmd, req->buffer, req->sector,
					       req->current_nr_sectors);
		}

	end_request:
		spin_lock_irq(&io_request_lock);
		end_request(retval);
	}
}

static volatile int out = 0;
static DECLARE_MUTEX_LOCKED(mpio_thread_sem);
static DECLARE_WAIT_QUEUE_HEAD(mpio_thread_wq);

int mpioblock_thread(void *data) {
	struct task_struct *task = current;
	DECLARE_WAITQUEUE(wait, task);

	task->pgrp = 1;
	task->tty = NULL;
	task->session = 1;
	task->flags |= PF_MEMALLOC;
	strcpy(task->comm, THREAD_NAME);

	spin_lock_irq(&task->sigmask_lock);
	sigfillset(&task->blocked);
	recalc_sigpending(task);
	spin_unlock_irq(&task->sigmask_lock);

	exit_mm(task);
	exit_files(task);
	exit_sighand(task);
	exit_fs(task);

	while (!out) {
		add_wait_queue(&mpio_thread_wq, &wait);
		set_current_state(TASK_INTERRUPTIBLE);
		spin_lock_irq(&io_request_lock);

		if (QUEUE_EMPTY || QUEUE_PLUGGED) {
			spin_unlock_irq(&io_request_lock);
			schedule();
			remove_wait_queue(&mpio_thread_wq, &wait);
		} else {
			remove_wait_queue(&mpio_thread_wq, &wait);
			set_current_state(TASK_RUNNING);
			mpioblock_handle_request();
			spin_unlock_irq(&io_request_lock);
		}
	}
	
	up(&mpio_thread_sem);
	return 0;
}

static void mpioblock_request(request_queue_t *q) {
	wake_up(&mpio_thread_wq);
}

static int __init mpioblock_init(void) {
	int i, retval;

#ifdef CONFIG_DEVFS_FS
	if (devfs_register_blkdev(MPIO_BLOCK_MAJOR, DEVICE_NAME, &mpio_fops)) {
		err("Can't allocate major number %d for MPIO.", MPIO_BLOCK_MAJOR);
		return -EAGAIN;
	}

	mpio_devfs_handle = devfs_mk_dir(NULL, DEVICE_NAME, NULL);
#else
	if (register_blkdev(MAJOR_NR, DEVICE_NAME, &mpio_fops)) {
		err("Can't allocate major number %d for MPIO.", MPIO_BLOCK_MAJOR);
		return -EAGAIN;
	}
#endif
	
	for (i = 0; i < MPIO_MAX_DEVICES; i++) {
		mpio_sizes[i] = 0;
		mpio_blksizes[i] = BLOCK_SIZE;
	}

	init_waitqueue_head(&mpio_thread_wq);

	blk_size[MAJOR_NR] = mpio_sizes;
	blksize_size[MAJOR_NR] = mpio_blksizes;

	/* USB driver should be registered after mpio_devfs_handle is
	 * initialized, because if usb device is attached to bus, then call to
	 * usb_register function will lead to requesting the mpioblock_probe
	 * method, which uses mpio_devfs_handle for creating minor devices in
	 * /dev/mpio */
	
	if ((retval = usb_register(&mpioblock_driver)) < 0) {
		err("Can't register mpio block driver.");
		return retval;
	}

	blk_init_queue(BLK_DEFAULT_QUEUE(MAJOR_NR), &mpioblock_request);
	kernel_thread(mpioblock_thread, NULL, CLONE_FS | CLONE_FILES | CLONE_SIGHAND );
		
	info(DRIVER_DESC " " DRIVER_VERSION " by " DRIVER_AUTHOR);
	return 0;
}

static void __exit mpioblock_exit(void) {
	
	out = 1;
	wake_up(&mpio_thread_wq);
	down(&mpio_thread_sem);
	
	usb_deregister(&mpioblock_driver);
	
#ifdef CONFIG_DEVFS_FS
	devfs_unregister(mpio_devfs_handle);
	devfs_unregister_blkdev(MPIO_BLOCK_MAJOR, DEVICE_NAME);
#else
	unregister_blkdev(MAJOR_NR, DEVICE_NAME);
#endif
	blk_cleanup_queue(BLK_DEFAULT_QUEUE(MAJOR_NR));

	blk_size[MAJOR_NR] = NULL;
	blksize_size[MAJOR_NR] = NULL;
}

module_init(mpioblock_init);
module_exit(mpioblock_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
