/*
  MPIO block device driver, version 0.0.1
  Copyright (C) 2002 Yury Umanets (umka@namesys.com)

  This program is free software; you can redistribute it and/or modify it under
  the terms of the GNU General Public License as published by the Free Software
  Foundation, version 2.
*/

#include <linux/config.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/usb.h>

#include "mpio.h"
#include "device.h"

static void mpio_completion(struct urb *urb) {
	if (urb->status != -ENOENT && urb->status != -ECONNRESET) {
		printk(KERN_DEBUG, __FUNCTION__ " - non-zero write bulk "
		       "status received: %d", urb->status);
	}
}

static int mpio_packet(struct mpio *mpio, struct mpio_packet *packet,
		       mpio_ctl_t ctl, mpio_stor_t stor, int sector, int size)
{
	/* Not supported yet */
	if (stor == STR_EXTERNAL)
		return -ENOSYS;

	memset(packet, 0, 0x40);

	/* Preparing packet */
	*packet->body = ctl;

	/* Check if we need to do bigendian translation, not supported yet */
	*(packet->body + 0x01) = stor;
	*(packet->body + 0x03) = (unsigned char)(sector & 0x00ff);
	*(packet->body + 0x04) = (unsigned char)((sector & 0xff00) >> 8);

	if (mpio->internal.size <= 32)
		*(packet->body + 0x05) = 0xff;
	else
		*(packet->body + 0x05) = (unsigned char)(sector >> 16);
	
	*(packet->body + 0x06) = size;
	memcpy(packet->body + 0x3b, "jykim", 5);
	
	return 0;
}

static char *models[MPIO_LAST] = {
	"MPIO DME",
	"MPIO DMG",
	"MPIO DMG+",
	"MPIO DMB",
	"MPIO DMB+",
	"MPIO DMK",
	"UNKNOWN"
};

static mpio_model_t mpio_model(struct mpio_packet *packet) {
	mpio_model_t model;
	
	switch (packet->body[5]) {
	case 'E':
		model = MPIO_DME;
		break;
	case 'K':
		model = MPIO_DMK;
		break;
	case 'G':
		model = MPIO_DMG;
		if (packet->body[6] == 'P')
			model = MPIO_DMG_PLUS;
		break;
	case 'B':
		model = MPIO_DMB;
		if (packet->body[6] == 'P')
			model = MPIO_DMB_PLUS;
		break;
	default:
		model = MPIO_UNKNOWN;
	}

	return model;
}

static int mpio_size(struct mpio_packet *packet) {
	int size;
	
	switch (packet->body[0x18 + 1]) {
	case 0xea:
		size = 2;
		break;
	case 0xe3:
	case 0xe5:
		size = 4;
		break;
	case 0xe6:
		size = 8;
		break;
	case 0x73:
		size = 16;
		break;
	case 0x75:
		size = 32;
		break;
	case 0x76:
		size = 64;
		break;
	case 0x79:
		size = 128;
		break;
	default:
		size = 0;
	}
	
	return size;
}

struct mpio *mpio_create(struct usb_device *udev, int ifnum) {
	struct mpio *mpio;
	struct usb_interface *interface;
	
	if (!(mpio = kmalloc(sizeof(struct mpio), GFP_KERNEL)))
		return NULL;

	memset(mpio, 0, sizeof(*mpio));

	if (!(mpio->ibuff = kmalloc(TRANS_SIZE*32, GFP_KERNEL)))
		goto error_free_mpio;

	if (!(mpio->obuff = kmalloc(TRANS_SIZE, GFP_KERNEL)))
		goto error_free_ibuff;

	mpio->size = 0xff;
	mpio->internal.size = 0xff;

	/* Initializing the mpio usb interface */
	interface = &udev->actconfig->interface[ifnum];
	
	mpio->iface.udev = udev;
	mpio->iface.interface = interface;

	init_waitqueue_head(&mpio->wq);

	return mpio;
	
 error_free_ibuff:
	kfree(mpio->ibuff);
 error_free_mpio:
	kfree(mpio);
	return NULL;
}

void mpio_free(struct mpio *mpio) {
	kfree(mpio->ibuff);
	kfree(mpio->obuff);
	kfree(mpio);
}

/* Makes request to device in odrer to initialize card sizes, etc */
int mpio_init(struct mpio *mpio) {
	int retval;
	char *model;
	unsigned int pipe, size;
	struct mpio_packet packet;

	if (mpio->initialized)
		return 0;
	
	memset(&packet, 0, sizeof(packet));
	
	if ((retval = mpio_packet(mpio, &packet, CTL_GET_VERSION, 0, 0, 0))) {
		err("Can't initialize init request.");
		return retval;
	}

	pipe = usb_sndbulkpipe(mpio->iface.udev, 1);
	if ((retval = usb_bulk_msg(mpio->iface.udev, pipe, packet.body,
				   0x40, &size, HZ*5)))
	{
		err("Can't send init packet.");
		return retval;
	}

	memset(&packet, 0, sizeof(packet));
	
	pipe = usb_rcvbulkpipe(mpio->iface.udev, 2);
	if ((retval = usb_bulk_msg(mpio->iface.udev, pipe, packet.body,
				   0x40, &size, HZ*5)))
	{
		err("Can't receive init packet.");
		return retval;
	}

	mpio->internal.size = mpio_size(&packet);

	if (!(mpio->size = mpio->internal.size*1024*1024)) {
		err("Invalid device size detected.");
		return -EINVAL;
	}

	mpio->model = mpio_model(&packet);
	mpio->initialized = 1;
	
	model = models[mpio->model];
	
	memcpy(&mpio->version, &packet, sizeof(packet));
	info("%s %d has been detected.", model, mpio->internal.size);

	return 0;
}

/* Shutdown any bulk writes that might be going on */
void mpio_fini(struct mpio *mpio) {
	
}

/* Reads data from passed MPIO */
int mpio_read(struct mpio *mpio, void *buff, int sector) {
	unsigned int pipe, size;
	struct mpio_packet packet;
	int retval, to_read, max_retry = 5;

	if (sector >= mpio->pstart &&
	    sector < mpio->pstart + mpio->pcount)
	{
		int offset = (sector - mpio->pstart) * SECTOR_SIZE;
		memcpy(buff, mpio->ibuff + offset, SECTOR_SIZE);
		return 0;
	}

	memset(&packet, 0, sizeof(packet));

	if ((retval = mpio_packet(mpio, &packet, CTL_GET_BLOCK,
				  STR_INTERNAL, sector, 0)))
	{
		err("Can't initialize read request.");
		return retval;
	}

	pipe = usb_sndbulkpipe(mpio->iface.udev, 1);
	if ((retval = usb_bulk_msg(mpio->iface.udev, pipe, packet.body,
				   0x40, &size, HZ*5)))
	{
		err("Can't send read sector request.");
		return retval;
	}

	if (size != 0x40)
		return -EIO;

	size = 0;
	
	for (to_read = TRANS_SIZE*32; to_read > 0; ) {
		
		if (signal_pending(current))
			return -EINTR;
		
		pipe = usb_rcvbulkpipe(mpio->iface.udev, 2);

		retval = usb_bulk_msg(mpio->iface.udev, pipe,
				      mpio->ibuff + size,
				      to_read, &size, HZ*5);

		if (size > 0) {
			to_read -= size;
		} else if (retval == USB_ST_TIMEOUT || retval == 15) {
			if (!max_retry--) {
				err("Timeout retry limit reached.");
				return -ETIME;
			}
			
			interruptible_sleep_on_timeout(&mpio->wq, HZ);
			continue;
			
		} else {
			if (retval != USB_ST_DATAUNDERRUN)
				err("Can't read data from MPIO.");
			
			return -EIO;
		}
	}
	
	memcpy(buff, mpio->ibuff, SECTOR_SIZE);

	mpio->pstart = sector;
	mpio->pcount = 32;

	return 0;
}

int mpio_write(struct mpio *mpio, void *buff, int sector) {
	unsigned int pipe, size;
	struct mpio_packet packet;
	int retval, to_write, max_retry = 5;

	memset(&packet, 0, sizeof(packet));

	if ((retval = mpio_packet(mpio, &packet, CTL_PUT_SECTOR,
				  STR_INTERNAL, sector, 0)))
	{
		err("Can't initialize write request.");
		return retval;
	}

	memset(mpio->obuff, 0, TRANS_SIZE);
	memcpy(mpio->obuff, buff, SECTOR_SIZE);
	
	pipe = usb_sndbulkpipe(mpio->iface.udev, 1);
	if ((retval = usb_bulk_msg(mpio->iface.udev, pipe, packet.body,
				   0x40, &size, HZ*5)))
	{
		err("Can't send write sector request.");
		return retval;
	}

	if (size != 0x40)
		return -EIO;
	
	size = 0;
	
	for (to_write = TRANS_SIZE; to_write > 0; ) {
		
		if (signal_pending(current))
			return -EINTR;
		
		pipe = usb_sndbulkpipe(mpio->iface.udev, 1);

		retval = usb_bulk_msg(mpio->iface.udev, pipe,
				      mpio->obuff + size,
				      to_write, &size, HZ*5);

		if (retval == USB_ST_TIMEOUT) {
			
			if (!max_retry--) {
				err("Timeout retry limit reached.");
				return -ETIME;
			}
			
			interruptible_sleep_on_timeout(&mpio->wq, HZ);
			continue;
			
		} else
			to_write -= size;
	}

	return 0;
}
